import pytest

from zapros import URL
from zapros._errors import TooManyRedirectsError
from zapros._handlers._mock import (
    Mock,
    MockMiddleware,
    MockRouter,
)
from zapros._handlers._redirect import (
    RedirectMiddleware,
)
from zapros._models import (
    Headers,
    Request,
    Response,
)
from zapros.matchers import host, path


def create_recording_handler():
    requests: list[Request] = []

    def record_request(
        request: Request,
    ) -> Response:
        requests.append(request)
        return Response(
            status=404,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    Mock().callback(record_request).mount(router)

    class RecordingHandlerWrapper:
        def __init__(self):
            self.requests = requests
            self.handler = MockMiddleware(router)

        def handle(self, request: Request) -> Response:
            return self.handler.handle(request)

        async def ahandle(self, request: Request) -> Response:
            return await self.handler.ahandle(request)

        def close(self) -> None:
            self.handler.close()

        async def aclose(self) -> None:
            await self.handler.aclose()

    return RecordingHandlerWrapper()


def test_redirect_301_preserves_method():
    router = MockRouter()

    Mock.given(path("/initial")).respond(
        Response(
            status=301,
            headers={"Location": "/final"},
        )
    ).mount(router)

    Mock.given(path("/final")).respond(
        Response(
            status=200,
            text="OK",
        )
    ).mount(router)

    mock_handler = MockMiddleware(router)
    handler = RedirectMiddleware(mock_handler)

    request = Request(
        URL("https://example.com/initial"),
        "POST",
        text="body content",
    )
    response = handler.handle(request)

    assert response.status == 200


def test_redirect_301_follows():
    router = MockRouter()
    Mock.given(path("/initial")).respond(
        Response(
            status=301,
            headers={"Location": "/final"},
        )
    ).mount(router)
    Mock.given(path("/final")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)

    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert response.text == "OK"


def test_redirect_302_follows():
    router = MockRouter()
    Mock.given(path("/initial")).respond(
        Response(
            status=302,
            headers={"Location": "/final"},
        )
    ).mount(router)
    Mock.given(path("/final")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert response.text == "OK"


def test_redirect_307_preserves_method():
    router = MockRouter()
    Mock.given(path("/initial")).respond(
        Response(
            status=307,
            headers={"Location": "/final"},
        )
    ).mount(router)
    Mock.given(path("/final")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "POST",
        text="test body",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    router.mocks[1]


def test_redirect_308_preserves_method():
    router = MockRouter()
    Mock.given(path("/initial")).respond(
        Response(
            status=308,
            headers={"Location": "/final"},
        )
    ).mount(router)
    Mock.given(path("/final")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "PUT",
        text="test body",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_301_converts_post_to_get():
    router = MockRouter()
    Mock.given(path("/initial")).respond(
        Response(
            status=301,
            headers={"Location": "/final"},
        )
    ).mount(router)
    Mock.given(path("/final")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "POST",
        text="test body",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_302_converts_post_to_get():
    router = MockRouter()
    Mock.given(path("/initial")).respond(
        Response(
            status=302,
            headers={"Location": "/final"},
        )
    ).mount(router)
    Mock.given(path("/final")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "POST",
        text="test body",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_303_converts_to_get():
    router = MockRouter()
    Mock.given(path("/initial")).respond(
        Response(
            status=303,
            headers={"Location": "/final"},
        )
    ).mount(router)
    Mock.given(path("/final")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "POST",
        text="test body",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_no_location_header():
    router = MockRouter()
    Mock.given(path("/initial")).respond(Response(status=301)).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 301


def test_redirect_max_redirects():
    router = MockRouter()
    for i in range(15):
        Mock.given(path(f"/step{i}")).respond(
            Response(
                status=302,
                headers={"Location": f"/step{i + 1}"},
            )
        ).mount(router)
    Mock.given(path("/step15")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler, max_redirects=5)

    request = Request(
        URL("https://example.com/step0"),
        "GET",
    )
    with pytest.raises(TooManyRedirectsError):
        redirect_handler.handle(request)


def test_redirect_updates_host_header():
    router = MockRouter()
    Mock.given(path("/initial")).mount(router).respond(
        Response(
            status=302,
            headers={"Location": "https://other.com/final"},
        )
    )
    Mock.given(host("other.com")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "GET",
        headers=Headers({"Host": "example.com"}),
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_strips_authorization():
    router = MockRouter()
    Mock.given(path("/initial")).respond(
        Response(
            status=302,
            headers={"Location": "/final"},
        )
    ).mount(router)
    Mock.given(path("/final")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "GET",
        headers=Headers({"Authorization": "Bearer token"}),
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_chain():
    router = MockRouter()
    Mock.given(path("/first")).respond(
        Response(
            status=302,
            headers={"Location": "/second"},
        )
    ).mount(router)
    Mock.given(path("/second")).respond(
        Response(
            status=302,
            headers={"Location": "/third"},
        )
    ).mount(router)
    Mock.given(path("/third")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/first"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert response.text == "OK"


def test_redirect_preserves_query_string():
    router = MockRouter()
    (
        Mock.given(path("/initial"))
        .respond(
            Response(
                status=302,
                headers={"Location": "/final?redirected=true"},
            )
        )
        .mount(router)
    )
    Mock.given(path("/final")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial?foo=bar"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


@pytest.mark.asyncio
async def test_async_redirect():
    router = MockRouter()
    Mock.given(path("/initial")).respond(
        Response(
            status=302,
            headers={"Location": "/final"},
        )
    ).mount(router)
    Mock.given(path("/final")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "GET",
    )
    response = await redirect_handler.ahandle(request)

    assert response.status == 200
    assert response.text == "OK"


@pytest.mark.asyncio
async def test_async_redirect_307_preserves_method():
    router = MockRouter()
    Mock.given(path("/initial")).respond(
        Response(
            status=307,
            headers={"Location": "/final"},
        )
    ).mount(router)
    Mock.given(path("/final")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "POST",
        text="body",
    )
    response = await redirect_handler.ahandle(request)

    assert response.status == 200


def test_redirect_non_redirect_status_passes_through():
    router = MockRouter()
    Mock.given(path("/")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert response.text == "OK"


def test_redirect_with_custom_port():
    router = MockRouter()
    Mock.given(path("/initial")).respond(
        Response(
            status=302,
            headers={"Location": "/final"},
        )
    ).mount(router)
    Mock.given(path("/final")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com:8080/initial"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_absolute_url():
    router = MockRouter()
    (
        Mock.given(path("/initial"))
        .respond(
            Response(
                status=302,
                headers={"Location": "https://other.example.com/final"},
            )
        )
        .mount(router)
    )
    Mock.given(path("/final")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_head_method_preserved():
    router = MockRouter()
    Mock.given(path("/initial")).respond(
        Response(
            status=302,
            headers={"Location": "/final"},
        )
    ).mount(router)
    Mock.given(path("/final")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "HEAD",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_delete_preserved_on_307():
    router = MockRouter()
    (
        Mock.given(path("/resource"))
        .respond(
            Response(
                status=307,
                headers={"Location": "/new-location"},
            )
        )
        .mount(router)
    )
    Mock.given(path("/new-location")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/resource"),
        "DELETE",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_delete_preserved_on_301():
    router = MockRouter()
    Mock.given(path("/resource")).respond(
        Response(
            status=301,
            headers={"Location": "/new-location"},
        )
    ).mount(router)
    Mock.given(path("/new-location")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/resource"),
        "DELETE",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_relative_path():
    router = MockRouter()
    Mock.given(path("/api/v1/resource")).respond(
        Response(
            status=302,
            headers={"Location": "../v2/resource"},
        )
    ).mount(router)
    Mock.given(path("/api/v2/resource")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/api/v1/resource"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_query_only():
    call_count = {"value": 0}

    def handle_with_counter(
        request: Request,
    ) -> Response:
        call_count["value"] += 1
        if call_count["value"] == 1:
            return Response(
                status=302,
                headers=Headers({"Location": "?page=2"}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    Mock().callback(handle_with_counter).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/search?page=1"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert call_count["value"] == 2


def test_redirect_fragment_only():
    call_count = {"value": 0}

    def handle_with_counter(
        request: Request,
    ) -> Response:
        call_count["value"] += 1
        if call_count["value"] == 1:
            return Response(
                status=302,
                headers=Headers({"Location": "#section"}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    Mock().callback(handle_with_counter).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/page"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert call_count["value"] == 2


def test_redirect_complex_relative():
    router = MockRouter()
    Mock.given(path("/docs/old")).respond(
        Response(
            status=302,
            headers={"Location": "./new/file"},
        )
    ).mount(router)
    Mock.given(path("/docs/new/file")).respond(Response(status=200, text="OK")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/docs/old"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_fragment_inheritance():
    router = MockRouter()
    Mock().respond(
        Response(
            status=302,
            headers={"Location": "/new-page"},
        )
    ).once().mount(router)
    Mock().given(path("/new-page")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/page#old"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_fragment_override():
    router = MockRouter()
    Mock().respond(
        Response(
            status=302,
            headers={"Location": "/new-page#new"},
        )
    ).once().mount(router)
    Mock().given(path("/new-page")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/page#old"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_fragment_none():
    router = MockRouter()
    Mock().respond(
        Response(
            status=302,
            headers={"Location": "/new-page"},
        )
    ).once().mount(router)
    Mock().given(path("/new-page")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/page"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_fragment_to_empty():
    router = MockRouter()
    Mock().respond(
        Response(
            status=302,
            headers={"Location": "/new-page#"},
        )
    ).once().mount(router)
    Mock().given(path("/new-page")).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/page#old"),
        "GET",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_301_put_preserved():
    router = MockRouter()
    Mock.given(path("/initial")).respond(
        Response(
            status=301,
            headers={"Location": "/final"},
        )
    ).mount(router)
    Mock.given(path("/final")).respond(Response(status=200)).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "PUT",
        text="test body",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200


def test_redirect_302_delete_preserved():
    methods = []

    def handle_with_method_tracking(
        request: Request,
    ) -> Response:
        methods.append(request.method)
        if "/initial" in str(request.url):
            return Response(
                status=302,
                headers=Headers({"Location": "/final"}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    Mock().callback(handle_with_method_tracking).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "DELETE",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert methods == [
        "DELETE",
        "DELETE",
    ]


def test_redirect_303_delete_to_get():
    methods = []

    def handle_with_method_tracking(
        request: Request,
    ) -> Response:
        methods.append(request.method)
        if "/initial" in str(request.url):
            return Response(
                status=303,
                headers=Headers({"Location": "/final"}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    Mock().callback(handle_with_method_tracking).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "DELETE",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert methods == ["DELETE", "GET"]


def test_redirect_303_head_preserved():
    methods = []

    def handle_with_method_tracking(
        request: Request,
    ) -> Response:
        methods.append(request.method)
        if "/initial" in str(request.url):
            return Response(
                status=303,
                headers=Headers({"Location": "/final"}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    Mock().callback(handle_with_method_tracking).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "HEAD",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert methods == ["HEAD", "HEAD"]


def test_redirect_strips_cookie():
    headers_list = []

    def handle_with_header_tracking(
        request: Request,
    ) -> Response:
        headers_list.append(dict(request.headers.list()))
        if "/initial" in str(request.url):
            return Response(
                status=302,
                headers=Headers({"Location": "/final"}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    Mock().callback(handle_with_header_tracking).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "GET",
        headers=Headers({"Cookie": "session=abc123"}),
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert "Cookie" in headers_list[0]
    assert "Cookie" not in headers_list[1]


def test_redirect_strips_if_none_match():
    headers_list = []

    def handle_with_header_tracking(
        request: Request,
    ) -> Response:
        headers_list.append(dict(request.headers.list()))
        if "/initial" in str(request.url):
            return Response(
                status=302,
                headers=Headers({"Location": "/final"}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    Mock().callback(handle_with_header_tracking).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "GET",
        headers=Headers({"If-None-Match": '"abc123"'}),
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert "If-None-Match" in headers_list[0]
    assert "If-None-Match" not in headers_list[1]


def test_redirect_strips_origin():
    headers_list = []

    def handle_with_header_tracking(
        request: Request,
    ) -> Response:
        headers_list.append(dict(request.headers.list()))
        if "/initial" in str(request.url):
            return Response(
                status=302,
                headers=Headers({"Location": "/final"}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    Mock().callback(handle_with_header_tracking).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "GET",
        headers=Headers({"Origin": "https://example.com"}),
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert "Origin" in headers_list[0]
    assert "Origin" not in headers_list[1]


def test_redirect_strips_content_encoding_on_get():
    headers_list = []

    def handle_with_header_tracking(
        request: Request,
    ) -> Response:
        headers_list.append(dict(request.headers.list()))
        if "/initial" in str(request.url):
            return Response(
                status=303,
                headers=Headers({"Location": "/final"}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    Mock().callback(handle_with_header_tracking).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "POST",
        headers=Headers({"Content-Encoding": "gzip"}),
        text="body",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert "Content-Encoding" in headers_list[0]
    assert "Content-Encoding" not in headers_list[1]


def test_redirect_preserves_content_type_on_307():
    headers_list = []

    def handle_with_header_tracking(
        request: Request,
    ) -> Response:
        headers_list.append(dict(request.headers.list()))
        if "/initial" in str(request.url):
            return Response(
                status=307,
                headers=Headers({"Location": "/final"}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    Mock().callback(handle_with_header_tracking).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "POST",
        headers=Headers({"Content-Type": "application/json"}),
        text='{"key": "value"}',
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert "Content-Type" in headers_list[0]
    assert "Content-Type" in headers_list[1]


def test_redirect_preserves_custom_headers():
    headers_list = []

    def handle_with_header_tracking(
        request: Request,
    ) -> Response:
        headers_list.append(dict(request.headers.list()))
        if "/initial" in str(request.url):
            return Response(
                status=302,
                headers=Headers({"Location": "/final"}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    Mock().callback(handle_with_header_tracking).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "GET",
        headers=Headers(
            {
                "X-Custom-Header": "value",
                "X-Another": "test",
            }
        ),
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert "X-Custom-Header" in headers_list[0]
    assert "X-Custom-Header" in headers_list[1]
    assert "X-Another" in headers_list[0]
    assert "X-Another" in headers_list[1]


def test_redirect_303_removes_body():
    bodies = []

    def handle_with_body_tracking(
        request: Request,
    ) -> Response:
        bodies.append(request.body)
        if "/initial" in str(request.url):
            return Response(
                status=303,
                headers=Headers({"Location": "/final"}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    Mock().callback(handle_with_body_tracking).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "POST",
        text="test body",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert bodies[0] is not None
    assert bodies[1] is None


def test_redirect_307_preserves_body():
    bodies = []

    def handle_with_body_tracking(
        request: Request,
    ) -> Response:
        bodies.append(request.body)
        if "/initial" in str(request.url):
            return Response(
                status=307,
                headers=Headers({"Location": "/final"}),
                content=None,
            )
        return Response(
            status=200,
            headers=Headers({}),
            content=None,
        )

    router = MockRouter()
    Mock().callback(handle_with_body_tracking).mount(router)

    handler = MockMiddleware(router)
    redirect_handler = RedirectMiddleware(handler)

    request = Request(
        URL("https://example.com/initial"),
        "POST",
        text="test body",
    )
    response = redirect_handler.handle(request)

    assert response.status == 200
    assert bodies[0] is not None
    assert bodies[1] is not None



from pywhatwgurl import URL as URL, URLSearchParams as URLSearchParams

from zapros._handlers._cookies import (
    CookieHandler as CookieHandler,  # type: ignore[reportDeprecated]
    CookieMiddleware as CookieMiddleware,
)
from zapros._handlers._proxy import ProxyMiddleware as ProxyMiddleware
from zapros._handlers._redirect import (
    RedirectHandler as RedirectHandler,  # type: ignore[reportDeprecated]
    RedirectMiddleware as RedirectMiddleware,
)
from zapros._handlers._retries import (
    RetryHandler as RetryHandler,  # type: ignore[reportDeprecated]
    RetryMiddleware as RetryMiddleware,
)
from zapros._io._asyncio import AsyncIOStream as AsyncIOStream, AsyncIOTransport as AsyncIOTransport
from zapros._io._sync import SyncStream as SyncStream, SyncTransport as SyncTransport

from ._async_client import (
    AsyncClient as AsyncClient,
)
from ._decoders import (
    DecodingError as DecodingError,
)
from ._errors import (
    AsyncSyncMismatchError as AsyncSyncMismatchError,
    ConnectionError as ConnectionError,
    ConnectTimeoutError as ConnectTimeoutError,
    PoolTimeoutError as PoolTimeoutError,
    ReadTimeoutError as ReadTimeoutError,
    TimeoutError as TimeoutError,
    TooManyRedirectsError as TooManyRedirectsError,
    TotalTimeoutError as TotalTimeoutError,
    WriteTimeoutError as WriteTimeoutError,
    ZaprosError as ZaprosError,
)
from ._handlers._asgi import (
    AsgiHandler as AsgiHandler,
)
from ._handlers._async_base import (
    AsyncBaseHandler as AsyncBaseHandler,
    AsyncBaseMiddleware as AsyncBaseMiddleware,
)
from ._handlers._async_pyodide import (
    AsyncPyodideHandler as AsyncPyodideHandler,
)
from ._handlers._async_pyreqwest import (
    AsyncPyreqwestHandler as AsyncPyreqwestHandler,
)
from ._handlers._async_std import (
    AsyncStdNetworkHandler as AsyncStdNetworkHandler,
)
from ._handlers._caching import (
    CacheMiddleware as CacheMiddleware,
    CachingHandler as CachingHandler,  # type: ignore[reportDeprecated]
)
from ._handlers._cassette import (
    Cassette as Cassette,
    CassetteHandler as CassetteHandler,  # type: ignore[reportDeprecated]
    CassetteMiddleware as CassetteMiddleware,
    Modifier as Modifier,
    UnhandledRequestError as UnhandledRequestError,
)
from ._handlers._sync_base import (
    BaseHandler as BaseHandler,
    BaseMiddleware as BaseMiddleware,
)
from ._handlers._sync_pyreqwest import (
    PyreqwestHandler as PyreqwestHandler,
)
from ._handlers._sync_std import (
    StdNetworkHandler as StdNetworkHandler,
)
from ._io._base import (
    AsyncBaseNetworkStream as AsyncBaseNetworkStream,
    AsyncBaseTransport as AsyncBaseTransport,
    BaseNetworkStream as BaseNetworkStream,
    BaseTransport as BaseTransport,
)
from ._models import (
    AsyncClosableStream as AsyncClosableStream,
    AsyncStream as AsyncStream,
    ClosableStream as ClosableStream,
    Headers as Headers,
    Request as Request,
    RequestContext as RequestContext,
    Response as Response,
    ResponseContext as ResponseContext,
    Stream as Stream,
)
from ._multidict import (
    CIMultiDict as CIMultiDict,
)
from ._multipart import (
    Multipart as Multipart,
    Part as Part,
)
from ._sync_client import (
    Client as Client,
)



from __future__ import annotations

from collections.abc import (
    Iterator as ABCIterator,
)
from typing import TYPE_CHECKING

from .._models import (
    ClosableStream,
    Request,
    Response,
)
from ._sync_base import (
    BaseHandler,
)
from ._exc_map import map_connect_exceptions, map_read_exceptions

if TYPE_CHECKING:
    import sys
    from contextlib import AbstractContextManager

    if sys.version_info >= (3, 11):
        from pyreqwest.client import (
            SyncClientBuilder,
        )
        from pyreqwest.response import (
            SyncResponse as PyreqwestResponse,
        )
    else:
        if not TYPE_CHECKING:
            SyncClientBuilder = None
            PyreqwestResponse = None
else:
    try:
        from pyreqwest.client import (
            SyncClientBuilder,
        )
        from pyreqwest.response import (
            SyncResponse as PyreqwestResponse,
        )
    except ImportError:
        if not TYPE_CHECKING:
            SyncClientBuilder = None
            PyreqwestResponse = None


class PyreqwestStream(ClosableStream):
    def __init__(
        self,
        response: PyreqwestResponse,
        stream_request: AbstractContextManager[PyreqwestResponse],
    ) -> None:
        self._response = response
        self._stream_request = stream_request

    def __next__(self) -> bytes:
        with map_read_exceptions():
            chunk = self._response.body_reader.read_chunk()
            if chunk is None:
                raise StopIteration
            return chunk.to_bytes()

    def close(self) -> None:
        self._stream_request.__exit__(None, None, None)


class PyreqwestHandler(BaseHandler):
    def __init__(
        self,
        client: SyncClientBuilder | None = None,
    ) -> None:
        if SyncClientBuilder is None:  # type: ignore[reportPossiblyUnboundVariable]
            raise ImportError("pyreqwest is not installed. Install it with: pip install pyreqwest")
        self._builder = client
        self._client = None

    def _get_client(self):
        if self._client is not None:
            return self._client
        builder = SyncClientBuilder() if self._builder is None else self._builder
        builder = builder.gzip(False).deflate(False).brotli(False).zstd(False)
        self._client = builder.build()
        return self._client

    def handle(self, request: Request) -> Response:
        client = self._get_client()

        req_builder = client.request(
            request.method,
            str(request.url),
        )

        for (
            key,
            value,
        ) in request.headers.items():
            req_builder = req_builder.header(key, value)

        if isinstance(request.body, bytes):
            req_builder = req_builder.body_bytes(request.body)
        elif isinstance(
            request.body,
            ABCIterator,
        ):
            req_builder = req_builder.body_stream(request.body)

        stream_request = req_builder.build_streamed()

        with map_connect_exceptions():
            pyreqwest_response = stream_request.__enter__()

        return Response(
            status=pyreqwest_response.status,
            headers=pyreqwest_response.headers,
            content=PyreqwestStream(pyreqwest_response, stream_request),
        )

    def close(self) -> None:
        if self._client is not None:
            self._client.close()
            self._client = None



from types import TracebackType
from unittest.mock import patch

from zapros import (
    AsyncStdNetworkHandler,
    StdNetworkHandler,
)

from ._handlers._mock import (
    Mock,
    MockHandler,  # type: ignore[reportDeprecated]
    MockMiddleware,
    MockRouter,
)
from .matchers import (
    AndMatcher,
    HeaderMatcher,
    HostMatcher,
    JsonMatcher,
    Matcher,
    MethodMatcher,
    NotMatcher,
    OrMatcher,
    PathMatcher,
    QueryMatcher,
    and_,
    header,
    host,
    json,
    method,
    not_,
    or_,
    path,
    query,
)

__all__ = [
    "AndMatcher",
    "HeaderMatcher",
    "HostMatcher",
    "JsonMatcher",
    "Matcher",
    "MethodMatcher",
    "Mock",
    "MockHandler",
    "MockMiddleware",
    "MockRouter",
    "NotMatcher",
    "OrMatcher",
    "PathMatcher",
    "QueryMatcher",
    "and_",
    "header",
    "host",
    "json",
    "method",
    "mock_http",
    "not_",
    "or_",
    "path",
    "query",
]


class mock_http:
    def __init__(
        self,
        mock_handler: MockMiddleware | None = None,
    ) -> None:
        self._mock_handler = mock_handler if mock_handler is not None else MockMiddleware(router=MockRouter())
        self._router = self._mock_handler._router  # type: ignore

        self._std_patch = patch.object(
            StdNetworkHandler,
            "handle",
            self._mock_handler.handle,
        )
        self._async_std_patch = patch.object(
            AsyncStdNetworkHandler,
            "ahandle",
            self._mock_handler.ahandle,
        )
        self._patches = [
            self._std_patch,
            self._async_std_patch,
        ]

    def __enter__(self) -> MockRouter:
        for p in self._patches:
            p.__enter__()
        return self._router

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        for p in reversed(self._patches):
            p.__exit__(
                exc_type,
                exc_val,
                exc_tb,
            )
        if exc_type is None:
            self._router.verify()
        self._router.reset()



import asyncio
import builtins
import mimetypes
import os
import uuid
from collections.abc import (
    AsyncIterator as ABCAsyncIterator,
    Iterator as ABCIterator,
)
from typing import (
    IO,
    AsyncIterator,
    Iterator,
)

from typing_extensions import Self

from ._models import (
    AsyncClosableStream,
    AsyncStream,
    ClosableStream,
    Stream,
)


def _escape_quoted_string(
    value: str,
) -> str:
    return value.replace("\\", "\\\\").replace('"', '\\"')


class _FileStream(AsyncClosableStream, ClosableStream):
    _CHUNK_SIZE = 65536

    def __init__(
        self,
        path: str | os.PathLike[str],
    ) -> None:
        self._path = path
        self._handle: IO[bytes] | None = None

    def _ensure_open_sync(
        self,
    ) -> IO[bytes]:
        if self._handle is None:
            self._handle = builtins.open(self._path, "rb")
        return self._handle

    async def _ensure_open_async(
        self,
    ) -> IO[bytes]:
        if self._handle is not None:
            return self._handle

        self._handle = await asyncio.to_thread(
            builtins.open,
            self._path,
            "rb",
        )
        return self._handle

    def __iter__(
        self,
    ) -> Iterator[bytes]:
        return self

    def __next__(self) -> bytes:
        handle = self._ensure_open_sync()
        chunk = handle.read(self._CHUNK_SIZE)
        if not chunk:
            self.close()
            raise StopIteration
        return chunk

    def __aiter__(
        self,
    ) -> "_FileStream":
        return self

    async def __anext__(self) -> bytes:
        handle = await self._ensure_open_async()
        chunk = await asyncio.to_thread(
            handle.read,
            self._CHUNK_SIZE,
        )
        if not chunk:
            await self.aclose()
            raise StopAsyncIteration
        return chunk

    def close(self) -> None:
        if self._handle is not None:
            self._handle.close()
            self._handle = None

    async def aclose(self) -> None:
        if self._handle is not None:
            await asyncio.to_thread(self._handle.close)
            self._handle = None


class Part:
    __slots__ = (
        "content",
        "filename",
        "content_type",
    )

    def __init__(
        self,
        content: Iterator[bytes] | AsyncIterator[bytes] | Stream | AsyncStream | bytes,
    ) -> None:
        self.content = content
        self.filename: str | None = None
        self.content_type: str = "application/octet-stream"

    @classmethod
    def bytes(cls, data: bytes) -> Self:
        return cls(data)

    @classmethod
    def stream(
        cls,
        stream: Iterator[builtins.bytes] | Stream,
    ) -> Self:
        return cls(stream)

    @classmethod
    def async_stream(
        cls,
        stream: AsyncIterator[builtins.bytes] | AsyncStream,
    ) -> Self:
        return cls(stream)

    @classmethod
    def text(cls, text: str) -> Self:
        part = cls(text.encode("utf-8"))
        part.content_type = "text/plain; charset=utf-8"
        return part

    def file_name(self, name: str) -> Self:
        self.filename = name
        return self

    def mime_type(self, mime: str) -> Self:
        self.content_type = mime
        return self

    def close(self) -> None:
        if isinstance(self.content, ClosableStream):
            self.content.close()
        elif isinstance(
            self.content,
            ABCAsyncIterator,
        ):
            raise RuntimeError("Part content is an AsyncStream, use aclose() to close it")

    async def aclose(self) -> None:
        if isinstance(
            self.content,
            AsyncClosableStream,
        ):
            await self.content.aclose()
        elif isinstance(self.content, ABCIterator):
            raise RuntimeError("Part content is a Stream, use close() to close it")


def _generate_part_headers(
    name: str,
    filename: str | None,
    content_type: str,
) -> bytes:
    escaped_name = _escape_quoted_string(name)

    if filename:
        escaped_filename = _escape_quoted_string(filename)
        disposition = f'Content-Disposition: form-data; name="{escaped_name}"; filename="{escaped_filename}"\r\n'
    else:
        disposition = f'Content-Disposition: form-data; name="{escaped_name}"\r\n'

    headers = disposition.encode("utf-8")
    headers += f"Content-Type: {content_type}\r\n\r\n".encode("utf-8")
    return headers


class MultipartStream(ClosableStream, AsyncClosableStream):
    __slots__ = (
        "_parts",
        "_boundary",
        "_iterator",
        "_async_iterator",
    )

    def __init__(
        self,
        parts: list[tuple[str, Part]],
        boundary: str,
    ) -> None:
        self._parts = parts
        self._boundary = boundary
        self._iterator: Iterator[bytes] = self._generate_parts()
        self._async_iterator: AsyncIterator[bytes] = self._generate_async_parts()

    def _generate_parts(
        self,
    ) -> Iterator[bytes]:
        for name, part in self._parts:
            yield f"--{self._boundary}\r\n".encode("utf-8")
            yield _generate_part_headers(
                name,
                part.filename,
                part.content_type,
            )

            assert not isinstance(
                part.content,
                ABCAsyncIterator,
            ), "Part content must be a Stream or bytes"

            if isinstance(part.content, bytes):
                yield part.content
            else:
                for chunk in part.content:
                    yield chunk

            yield b"\r\n"

        yield f"--{self._boundary}--\r\n".encode("utf-8")

    async def _generate_async_parts(
        self,
    ) -> AsyncIterator[bytes]:
        for name, part in self._parts:
            yield f"--{self._boundary}\r\n".encode("utf-8")
            yield _generate_part_headers(
                name,
                part.filename,
                part.content_type,
            )

            assert not isinstance(
                part.content,
                ABCIterator,
            ), "Part content must be an AsyncStream or bytes"

            if isinstance(part.content, bytes):
                yield part.content
            else:
                async for chunk in part.content:
                    yield chunk

            yield b"\r\n"

        yield f"--{self._boundary}--\r\n".encode("utf-8")

    def __aiter__(self):
        return self

    async def __anext__(self) -> bytes:
        return await anext(self._async_iterator)

    def __iter__(self):
        return self

    def __next__(self) -> bytes:
        return next(self._iterator)

    def close(self) -> None:
        for _, part in self._parts:
            part.close()

    async def aclose(self) -> None:
        for _, part in self._parts:
            await part.aclose()


class Multipart:
    __slots__ = ("_parts", "_boundary")

    def __init__(
        self,
        boundary: str | None = None,
    ) -> None:
        self._boundary = boundary or f"----FormBoundary{uuid.uuid4().hex}"
        self._parts: list[tuple[str, Part]] = []

    def text(self, name: str, value: str) -> Self:
        part = Part(value.encode("utf-8"))
        part.content_type = "text/plain; charset=utf-8"
        self._parts.append((name, part))
        return self

    def file(
        self,
        name: str,
        path: str | os.PathLike[str],
    ) -> Self:
        part = Part(_FileStream(path))
        part.filename = os.path.basename(os.fspath(path))
        part.content_type = mimetypes.guess_type(os.fspath(path))[0] or "application/octet-stream"
        self._parts.append((name, part))
        return self

    def part(self, name: str, part: Part) -> Self:
        self._parts.append((name, part))
        return self

    def to_body(
        self,
    ) -> bytes | MultipartStream:
        all_static = all(isinstance(part.content, bytes) for _, part in self._parts)

        if all_static:
            result = bytearray()
            for (
                name,
                part,
            ) in self._parts:
                result.extend(f"--{self._boundary}\r\n".encode("utf-8"))
                result.extend(
                    _generate_part_headers(
                        name,
                        part.filename,
                        part.content_type,
                    )
                )
                assert isinstance(part.content, bytes)
                result.extend(part.content)
                result.extend(b"\r\n")
            result.extend(f"--{self._boundary}--\r\n".encode("utf-8"))
            return bytes(result)
        else:
            return MultipartStream(
                parts=self._parts,
                boundary=self._boundary,
            )

    def to_async_body(
        self,
    ) -> bytes | MultipartStream:
        all_static = all(isinstance(part.content, bytes) for _, part in self._parts)

        if all_static:
            result = bytearray()
            for (
                name,
                part,
            ) in self._parts:
                result.extend(f"--{self._boundary}\r\n".encode("utf-8"))
                result.extend(
                    _generate_part_headers(
                        name,
                        part.filename,
                        part.content_type,
                    )
                )
                assert isinstance(part.content, bytes)
                result.extend(part.content)
                result.extend(b"\r\n")
            result.extend(f"--{self._boundary}--\r\n".encode("utf-8"))
            return bytes(result)
        else:
            return MultipartStream(
                parts=self._parts,
                boundary=self._boundary,
            )

    @property
    def content_type(self) -> str:
        return f'multipart/form-data; boundary="{self._boundary}"'



from abc import abstractmethod

from typing_extensions import Protocol, runtime_checkable


@runtime_checkable
class BaseNetworkStream(Protocol):
    @abstractmethod
    def read(self, max_bytes: int, timeout: float | None = None) -> bytes: ...
    @abstractmethod
    def write_all(self, data: bytes, timeout: float | None = None) -> int: ...
    def close(self) -> None: ...
    def start_tls(
        self,
        *,
        server_hostname: str | None = None,
    ) -> "BaseNetworkStream":
        return self


@runtime_checkable
class AsyncBaseNetworkStream(Protocol):
    @abstractmethod
    async def read(self, max_bytes: int, timeout: float | None = None) -> bytes: ...
    @abstractmethod
    async def write_all(self, data: bytes, timeout: float | None = None) -> int: ...
    async def start_tls(
        self,
        *,
        server_hostname: str | None = None,
    ) -> "AsyncBaseNetworkStream":
        return self

    async def close(self) -> None: ...


class BaseTransport(Protocol):
    def connect(
        self,
        host: str,
        port: int,
        server_hostname: str | None = None,
        tls: bool = False,
        *,
        timeout: float | None = None,
    ) -> BaseNetworkStream: ...


class AsyncBaseTransport(Protocol):
    async def aconnect(
        self,
        host: str,
        port: int,
        server_hostname: str | None = None,
        tls: bool = False,
        *,
        timeout: float | None = None,
    ) -> AsyncBaseNetworkStream: ...



import socket
import ssl
from typing import Any, Callable, TypeVar

from zapros._constants import DEFAULT_READ_SIZE, DEFAULT_SSL_CONTEXT
from zapros._handlers._exc_map import map_connect_exceptions, map_read_exceptions, map_write_exceptions
from zapros._io._base import BaseNetworkStream, BaseTransport

T = TypeVar("T")


class _SyncTLSState:
    def __init__(
        self,
        sock: socket.socket,
        ssl_object: ssl.SSLObject,
        incoming_bio: ssl.MemoryBIO,
        outgoing_bio: ssl.MemoryBIO,
    ) -> None:
        self.sock = sock
        self.ssl_object = ssl_object
        self.incoming_bio = incoming_bio
        self.outgoing_bio = outgoing_bio
        self.handshake_complete = False


class SyncStream(BaseNetworkStream):
    def __init__(self, sock: socket.socket, *, upgrade_ssl_context: ssl.SSLContext) -> None:
        self.sock = sock
        self._closed = False
        self._upgrade_ssl_context = upgrade_ssl_context
        self._tls_state: _SyncTLSState | None = None

    def _flush_outgoing(self, timeout: float | None = None) -> None:
        assert self._tls_state is not None
        pending = self._tls_state.outgoing_bio.read(DEFAULT_READ_SIZE)
        if pending:
            with map_write_exceptions():
                self._tls_state.sock.settimeout(timeout)
                self._tls_state.sock.sendall(pending)

    def _pump_incoming(self, timeout: float | None = None) -> None:
        assert self._tls_state is not None
        with map_read_exceptions():
            self._tls_state.sock.settimeout(timeout)
            data = self._tls_state.sock.recv(DEFAULT_READ_SIZE)
            if data:
                self._tls_state.incoming_bio.write(data)
            else:
                self._tls_state.incoming_bio.write_eof()

    def _call_sslobject_method(self, func: Callable[..., T], *args: Any) -> T:
        assert self._tls_state is not None
        while True:
            try:
                result = func(*args)
                self._flush_outgoing()
                return result
            except ssl.SSLWantReadError:
                self._flush_outgoing()
                self._pump_incoming()
            except ssl.SSLWantWriteError:
                self._flush_outgoing()
            except (ssl.SSLError, ssl.SSLEOFError) as e:
                raise ConnectionError(str(e)) from e

    def read(self, max_bytes: int, timeout: float | None = None) -> bytes:
        if self._tls_state is not None:
            return self._call_sslobject_method(self._tls_state.ssl_object.read, max_bytes)

        with map_read_exceptions():
            self.sock.settimeout(timeout)
            return self.sock.recv(max_bytes)

    def write_all(self, data: bytes, timeout: float | None = None) -> int:
        if self._tls_state is not None:
            self._call_sslobject_method(self._tls_state.ssl_object.write, data)
            return len(data)

        with map_write_exceptions():
            self.sock.settimeout(timeout)
            self.sock.sendall(data)
            return len(data)

    def close(self) -> None:
        if self._closed:
            return
        self._closed = True
        try:
            try:
                self.sock.shutdown(socket.SHUT_RDWR)
            except Exception:
                pass
            self.sock.close()
        except Exception:
            pass

    def start_tls(self, *, server_hostname: str | None = None) -> BaseNetworkStream:
        assert self._upgrade_ssl_context is not None

        hostname = server_hostname

        incoming = ssl.MemoryBIO()
        outgoing = ssl.MemoryBIO()

        ssl_object = self._upgrade_ssl_context.wrap_bio(incoming, outgoing, server_side=False, server_hostname=hostname)

        self._tls_state = _SyncTLSState(self.sock, ssl_object, incoming, outgoing)

        self._call_sslobject_method(self._tls_state.ssl_object.do_handshake)
        self._tls_state.handshake_complete = True

        return self


class SyncTransport(BaseTransport):
    def __init__(
        self,
        *,
        ssl_context: ssl.SSLContext | None = None,
    ) -> None:
        self.ssl_context = DEFAULT_SSL_CONTEXT if ssl_context is None else ssl_context

    def connect(
        self,
        host: str,
        port: int,
        server_hostname: str | None = None,
        tls: bool = False,
        *,
        timeout: float | None = None,
    ) -> BaseNetworkStream:
        with map_connect_exceptions():
            sock = socket.create_connection((host, port), timeout=timeout)

        try:
            if tls:
                sock = self.ssl_context.wrap_socket(
                    sock,
                    server_hostname=server_hostname,
                )
            sock.settimeout(None)
            return SyncStream(sock, upgrade_ssl_context=self.ssl_context)
        except Exception:
            sock.close()
            raise



from __future__ import annotations

from abc import abstractmethod
from typing import (
    Callable,
    Protocol,
    runtime_checkable,
)

from .._models import Request, Response


@runtime_checkable
class BaseHandler(Protocol):
    @abstractmethod
    def handle(self, request: Request) -> Response:
        raise NotImplementedError()

    def close(self) -> None:
        pass

    def wrap_with_middleware(
        self,
        factory: Callable[
            [BaseHandler],
            BaseMiddleware,
        ],
    ) -> BaseMiddleware:
        return factory(self)


@runtime_checkable
class BaseMiddleware(BaseHandler, Protocol):
    next: BaseHandler



from __future__ import annotations


class ZaprosError(Exception):
    """Base class for all Zapros errors."""

    pass


class AsyncSyncMismatchError(ZaprosError):
    """Raised when an asynchronous handler or function is used in a synchronous context, or vice versa."""

    def __init__(self, message: str) -> None:
        super().__init__(
            f"{message}\nSee https://zapros.dev/guide/async-sync.html#asyncsyncmismatcherror for more info."
        )


class ConnectionError(ZaprosError):
    """Raised when a connection cannot be established or is lost."""

    pass


class DNSResolutionError(ConnectionError):
    """Raised when DNS resolution fails."""

    pass


class TimeoutError(ZaprosError):
    """Base class for all Zapros timeout errors."""

    pass


class ConnectTimeoutError(TimeoutError):
    """Raised when connection establishment times out."""

    pass


class ReadTimeoutError(TimeoutError):
    """Raised when reading response data times out."""

    pass


class WriteTimeoutError(TimeoutError):
    """Raised when writing request data times out."""

    pass


class TotalTimeoutError(TimeoutError):
    """Raised when the total request deadline is exceeded."""

    pass


class PoolTimeoutError(TimeoutError):
    """Raised when waiting for a connection pool slot times out."""

    pass


class TooManyRedirectsError(ZaprosError):
    """Raised when the maximum number of redirects is exceeded."""

    pass


class AuthenticationError(ZaprosError):
    """Base class for all authentication-related errors."""

    pass


class ResponseNotRead(ZaprosError):
    """Raised when attempting to access response body content that hasn't been read yet."""

    pass



import json as json_module
from abc import abstractmethod
from typing import (
    Any,
    Callable,
    Protocol,
    runtime_checkable,
)

from zapros._models import Request

__all__ = [
    "AndMatcher",
    "HeaderMatcher",
    "HostMatcher",
    "JsonMatcher",
    "Matcher",
    "MethodMatcher",
    "NotMatcher",
    "OrMatcher",
    "PathMatcher",
    "QueryMatcher",
    "and_",
    "header",
    "host",
    "json",
    "method",
    "not_",
    "or_",
    "path",
    "query",
]


@runtime_checkable
class Matcher(Protocol):
    @abstractmethod
    def match(self, request: Request) -> bool:
        raise NotImplementedError

    def and_(self, other: "Matcher") -> "AndMatcher":
        return AndMatcher(self, other)

    def or_(self, other: "Matcher") -> "OrMatcher":
        return OrMatcher(self, other)

    def method(self, method: str) -> "AndMatcher":
        return AndMatcher(self, MethodMatcher(method))

    def path(self, path: str) -> "AndMatcher":
        return AndMatcher(self, PathMatcher(path))

    def host(self, host: str) -> "AndMatcher":
        return AndMatcher(self, HostMatcher(host))

    def header(self, name: str, value: str) -> "AndMatcher":
        return AndMatcher(
            self,
            HeaderMatcher(name, value),
        )

    def query(self, **params: str) -> "AndMatcher":
        return AndMatcher(self, QueryMatcher(**params))

    def json(
        self,
        predicate: Callable[[Any], bool],
    ) -> "AndMatcher":
        return AndMatcher(self, JsonMatcher(predicate))


class AndMatcher(Matcher):
    def __init__(self, *matchers: Matcher) -> None:
        self.matchers = matchers

    def match(self, request: Request) -> bool:
        return all(m.match(request) for m in self.matchers)


class OrMatcher(Matcher):
    def __init__(self, *matchers: Matcher) -> None:
        self.matchers = matchers

    def match(self, request: Request) -> bool:
        return any(m.match(request) for m in self.matchers)


class MethodMatcher(Matcher):
    def __init__(self, method: str) -> None:
        self._method = method.upper()

    def match(self, request: Request) -> bool:
        return request.method.upper() == self._method


class PathMatcher(Matcher):
    def __init__(self, path: str) -> None:
        self._path = path

    def match(self, request: Request) -> bool:
        return request.url.pathname == self._path


class HostMatcher(Matcher):
    def __init__(self, host: str) -> None:
        self._host = host

    def match(self, request: Request) -> bool:
        return request.url.hostname == self._host


class HeaderMatcher(Matcher):
    def __init__(self, name: str, value: str) -> None:
        self._name = name
        self._value = value

    def match(self, request: Request) -> bool:
        return request.headers.get(self._name) == self._value


class QueryMatcher(Matcher):
    def __init__(self, **params: str) -> None:
        self._params = params

    def match(self, request: Request) -> bool:
        query_dict = request.url.search_params
        return all(query_dict.get(key) == value for key, value in self._params.items())


class JsonMatcher(Matcher):
    def __init__(
        self,
        predicate: Callable[[Any], bool],
    ) -> None:
        self._predicate = predicate

    def match(self, request: Request) -> bool:
        if not isinstance(request.body, bytes):
            return False
        try:
            data = json_module.loads(request.body.decode("utf-8"))
            return self._predicate(data)
        except (
            ValueError,
            UnicodeDecodeError,
        ):
            return False


class NotMatcher(Matcher):
    def __init__(self, matcher: Matcher) -> None:
        self.matcher = matcher

    def match(self, request: Request) -> bool:
        return not self.matcher.match(request)


def and_(
    *matchers: Matcher,
) -> AndMatcher:
    return AndMatcher(*matchers)


def or_(
    *matchers: Matcher,
) -> OrMatcher:
    return OrMatcher(*matchers)


def not_(
    matcher: Matcher,
) -> NotMatcher:
    return NotMatcher(matcher)


def method(
    method: str,
) -> MethodMatcher:
    return MethodMatcher(method)


def path(path: str) -> PathMatcher:
    return PathMatcher(path)


def host(host: str) -> HostMatcher:
    return HostMatcher(host)


def header(name: str, value: str) -> HeaderMatcher:
    return HeaderMatcher(name, value)


def query(
    **params: str,
) -> QueryMatcher:
    return QueryMatcher(**params)


def json(
    predicate: Callable[[Any], bool],
) -> JsonMatcher:
    return JsonMatcher(predicate)



import gzip
import json as json_module
import zlib
from typing import AsyncIterator, Iterator

import pytest
from pywhatwgurl import URL

from zapros import (
    AsyncClosableStream,
    AsyncSyncMismatchError,
    ClosableStream,
    Headers,
    Request,
    RequestContext,
    Response,
    ResponseContext,
)


class StreamWrapper(ClosableStream):
    def __init__(self, iterator: Iterator[bytes]) -> None:
        self._iterator = iterator
        self._closed = False

    def __iter__(self) -> "StreamWrapper":
        return self

    def __next__(self) -> bytes:
        return next(self._iterator)

    def close(self) -> None:
        self._closed = True


class AsyncStreamWrapper(AsyncClosableStream):
    def __init__(self, generator: AsyncIterator[bytes]) -> None:
        self._generator = generator
        self._closed = False

    def __aiter__(self) -> "AsyncStreamWrapper":
        return self

    async def __anext__(self) -> bytes:
        return await self._generator.__anext__()

    async def aclose(self) -> None:
        self._closed = True


class TestHeaders:
    def test_headers_construction_empty(self):
        headers = Headers()
        assert len(headers) == 0

    def test_headers_construction_from_dict(self):
        headers = Headers({"Content-Type": "application/json"})
        assert headers["Content-Type"] == "application/json"
        assert len(headers) == 1

    def test_headers_construction_from_list_of_tuples(self):
        headers = Headers([("Content-Type", "application/json"), ("Accept", "*/*")])
        assert headers["Content-Type"] == "application/json"
        assert headers["Accept"] == "*/*"
        assert len(headers) == 2

    def test_headers_construction_from_headers_instance(self):
        original = Headers({"Content-Type": "application/json"})
        headers = Headers(original)
        assert headers["Content-Type"] == "application/json"
        assert len(headers) == 1

    def test_headers_getitem(self):
        headers = Headers({"Content-Type": "application/json"})
        assert headers["Content-Type"] == "application/json"

    def test_headers_getitem_case_insensitive(self):
        headers = Headers({"Content-Type": "application/json"})
        assert headers["content-type"] == "application/json"
        assert headers["CONTENT-TYPE"] == "application/json"

    def test_headers_getitem_missing_raises_keyerror(self):
        headers = Headers()
        with pytest.raises(KeyError):
            _ = headers["Missing-Header"]

    def test_headers_setitem(self):
        headers = Headers()
        headers["Content-Type"] = "application/json"
        assert headers["Content-Type"] == "application/json"

    def test_headers_setitem_case_insensitive(self):
        headers = Headers({"Content-Type": "application/json"})
        headers["content-type"] = "text/plain"
        assert headers["Content-Type"] == "text/plain"
        assert len(headers) == 1

    def test_headers_setitem_replaces_existing(self):
        headers = Headers({"Content-Type": "application/json"})
        headers["Content-Type"] = "text/plain"
        assert headers["Content-Type"] == "text/plain"

    def test_headers_delitem(self):
        headers = Headers({"Content-Type": "application/json"})
        del headers["Content-Type"]
        assert len(headers) == 0

    def test_headers_delitem_case_insensitive(self):
        headers = Headers({"Content-Type": "application/json"})
        del headers["content-type"]
        assert len(headers) == 0

    def test_headers_delitem_missing_raises_keyerror(self):
        headers = Headers()
        with pytest.raises(KeyError):
            del headers["Missing-Header"]

    def test_headers_contains(self):
        headers = Headers({"Content-Type": "application/json"})
        assert "Content-Type" in headers
        assert "Missing-Header" not in headers

    def test_headers_contains_case_insensitive(self):
        headers = Headers({"Content-Type": "application/json"})
        assert "content-type" in headers
        assert "CONTENT-TYPE" in headers

    def test_headers_iter(self):
        headers = Headers({"Content-Type": "application/json", "Accept": "*/*"})
        keys = list(headers)
        assert len(keys) == 2
        assert "Content-Type" in keys or "content-type" in keys
        assert "Accept" in keys or "accept" in keys

    def test_headers_len(self):
        headers = Headers()
        assert len(headers) == 0
        headers["Content-Type"] = "application/json"
        assert len(headers) == 1
        headers["Accept"] = "*/*"
        assert len(headers) == 2

    def test_headers_repr(self):
        headers = Headers({"Content-Type": "application/json"})
        repr_str = repr(headers)
        assert "Headers" in repr_str
        assert "Content-Type" in repr_str or "content-type" in repr_str

    def test_headers_get_with_default(self):
        headers = Headers({"Content-Type": "application/json"})
        assert headers.get("Content-Type") == "application/json"
        assert headers.get("Missing-Header") is None
        assert headers.get("Missing-Header", "default") == "default"

    def test_headers_get_case_insensitive(self):
        headers = Headers({"Content-Type": "application/json"})
        assert headers.get("content-type") == "application/json"

    def test_headers_getall(self):
        headers = Headers([("Set-Cookie", "a=1"), ("Set-Cookie", "b=2")])
        cookies = headers.getall("Set-Cookie")
        assert len(cookies) == 2
        assert "a=1" in cookies
        assert "b=2" in cookies

    def test_headers_getall_case_insensitive(self):
        headers = Headers([("Set-Cookie", "a=1"), ("Set-Cookie", "b=2")])
        cookies = headers.getall("set-cookie")
        assert len(cookies) == 2

    def test_headers_getall_single_value(self):
        headers = Headers({"Content-Type": "application/json"})
        values = headers.getall("Content-Type")
        assert values == ["application/json"]

    def test_headers_add(self):
        headers = Headers()
        headers.add("Set-Cookie", "a=1")
        headers.add("Set-Cookie", "b=2")
        cookies = headers.getall("Set-Cookie")
        assert len(cookies) == 2

    def test_headers_add_case_insensitive(self):
        headers = Headers()
        headers.add("Set-Cookie", "a=1")
        headers.add("set-cookie", "b=2")
        cookies = headers.getall("Set-Cookie")
        assert len(cookies) == 2

    def test_headers_extend_from_dict(self):
        headers = Headers()
        headers.extend({"Content-Type": "application/json", "Accept": "*/*"})
        assert len(headers) == 2

    def test_headers_extend_from_list(self):
        headers = Headers()
        headers.extend([("Set-Cookie", "a=1"), ("Set-Cookie", "b=2")])
        cookies = headers.getall("Set-Cookie")
        assert len(cookies) == 2

    def test_headers_pop_with_default(self):
        headers = Headers({"Content-Type": "application/json"})
        value = headers.pop("Content-Type")
        assert value == "application/json"
        assert len(headers) == 0

    def test_headers_pop_with_default_value(self):
        headers = Headers()
        value = headers.pop("Missing-Header", "default")
        assert value == "default"

    def test_headers_pop_case_insensitive(self):
        headers = Headers({"Content-Type": "application/json"})
        value = headers.pop("content-type")
        assert value == "application/json"

    def test_headers_popitem(self):
        headers = Headers({"Content-Type": "application/json"})
        key, value = headers.popitem()
        assert key.lower() == "content-type"
        assert value == "application/json"
        assert len(headers) == 0

    def test_headers_popitem_empty_raises_keyerror(self):
        headers = Headers()
        with pytest.raises(KeyError):
            headers.popitem()

    def test_headers_clear(self):
        headers = Headers({"Content-Type": "application/json", "Accept": "*/*"})
        headers.clear()
        assert len(headers) == 0

    def test_headers_setdefault_new_key(self):
        headers = Headers()
        value = headers.setdefault("Content-Type", "application/json")
        assert value == "application/json"
        assert headers["Content-Type"] == "application/json"

    def test_headers_setdefault_existing_key(self):
        headers = Headers({"Content-Type": "application/json"})
        value = headers.setdefault("Content-Type", "text/plain")
        assert value == "application/json"
        assert headers["Content-Type"] == "application/json"

    def test_headers_setdefault_case_insensitive(self):
        headers = Headers({"Content-Type": "application/json"})
        value = headers.setdefault("content-type", "text/plain")
        assert value == "application/json"

    def test_headers_update_from_dict(self):
        headers = Headers()
        headers.update({"Content-Type": "application/json", "Accept": "*/*"})
        assert len(headers) == 2

    def test_headers_update_from_list(self):
        headers = Headers()
        headers.update([("Content-Type", "application/json"), ("Accept", "*/*")])
        assert len(headers) == 2

    def test_headers_copy(self):
        headers = Headers({"Content-Type": "application/json"})
        copied = headers.copy()
        assert copied["Content-Type"] == "application/json"
        copied["Accept"] = "*/*"
        assert "Accept" not in headers

    def test_headers_list(self):
        headers = Headers({"Content-Type": "application/json", "Accept": "*/*"})
        header_list = headers.list()
        assert len(header_list) == 2
        assert any(k.lower() == "content-type" and v == "application/json" for k, v in header_list)
        assert any(k.lower() == "accept" and v == "*/*" for k, v in header_list)

    def test_headers_keys(self):
        headers = Headers({"Content-Type": "application/json", "Accept": "*/*"})
        keys = list(headers.keys())
        assert len(keys) == 2

    def test_headers_values(self):
        headers = Headers({"Content-Type": "application/json", "Accept": "*/*"})
        values = list(headers.values())
        assert len(values) == 2
        assert "application/json" in values
        assert "*/*" in values

    def test_headers_items(self):
        headers = Headers({"Content-Type": "application/json"})
        items = list(headers.items())
        assert len(items) == 1
        key, value = items[0]
        assert key.lower() == "content-type"
        assert value == "application/json"


class TestRequest:
    def test_request_minimal(self):
        url = URL("http://example.com/path")
        request = Request(url, "GET")
        assert request.url == url
        assert request.method == "GET"
        assert request.body is None
        assert request.context == {}

    def test_request_json_parameter(self):
        url = URL("http://example.com/path")
        data = {"key": "value"}
        request = Request(url, "POST", json=data)
        assert isinstance(request.body, bytes)
        assert json_module.loads(request.body) == data
        assert request.headers["Content-Type"] == "application/json"
        assert "Content-Length" in request.headers

    def test_request_json_sets_content_type(self):
        url = URL("http://example.com/path")
        request = Request(url, "POST", json={"key": "value"})
        assert request.headers["Content-Type"] == "application/json"

    def test_request_json_does_not_override_content_type(self):
        url = URL("http://example.com/path")
        request = Request(url, "POST", headers={"Content-Type": "custom"}, json={"key": "value"})
        assert request.headers["Content-Type"] == "custom"

    def test_request_form_parameter(self):
        url = URL("http://example.com/path")
        form_data = {"key": "value", "name": "test"}
        request = Request(url, "POST", form=form_data)
        assert isinstance(request.body, bytes)
        assert b"key=value" in request.body
        assert b"name=test" in request.body
        assert request.headers["Content-Type"] == "application/x-www-form-urlencoded"
        assert "Content-Length" in request.headers

    def test_request_form_does_not_override_content_type(self):
        url = URL("http://example.com/path")
        request = Request(url, "POST", headers={"Content-Type": "custom"}, form={"key": "value"})
        assert request.headers["Content-Type"] == "custom"

    def test_request_text_parameter(self):
        url = URL("http://example.com/path")
        text = "Hello, World!"
        request = Request(url, "POST", text=text)
        assert request.body == text.encode("utf-8")
        assert request.headers["Content-Type"] == "text/plain; charset=utf-8"
        assert "Content-Length" in request.headers

    def test_request_text_does_not_override_content_type(self):
        url = URL("http://example.com/path")
        request = Request(url, "POST", headers={"Content-Type": "custom"}, text="Hello")
        assert request.headers["Content-Type"] == "custom"

    def test_request_body_bytes(self):
        url = URL("http://example.com/path")
        body = b"raw bytes"
        request = Request(url, "POST", body=body)
        assert request.body == body
        assert request.headers["Content-Length"] == str(len(body))

    def test_request_body_bytes_does_not_override_content_length(self):
        url = URL("http://example.com/path")
        request = Request(url, "POST", headers={"Content-Length": "999"}, body=b"test")
        assert request.headers["Content-Length"] == "999"

    def test_request_body_iterator(self):
        url = URL("http://example.com/path")
        body = iter([b"chunk1", b"chunk2"])
        request = Request(url, "POST", body=body)
        assert request.body == body
        assert request.headers["Transfer-Encoding"] == "chunked"

    def test_request_body_iterator_does_not_override_transfer_encoding(self):
        url = URL("http://example.com/path")
        body = iter([b"chunk1"])
        request = Request(url, "POST", headers={"Transfer-Encoding": "custom"}, body=body)
        assert request.headers["Transfer-Encoding"] == "custom"

    def test_request_body_async_iterator(self):
        url = URL("http://example.com/path")

        async def body_gen():
            yield b"chunk1"

        request = Request(url, "POST", body=body_gen())
        assert request.headers["Transfer-Encoding"] == "chunked"

    def test_request_no_body_parameter(self):
        url = URL("http://example.com/path")
        request = Request(url, "GET")
        assert request.body is None
        assert "Content-Length" not in request.headers
        assert "Transfer-Encoding" not in request.headers

    def test_request_host_header_from_url(self):
        url = URL("http://example.com/path")
        request = Request(url, "GET")
        assert request.headers["Host"] == "example.com"

    def test_request_host_header_not_overridden(self):
        url = URL("http://example.com/path")
        request = Request(url, "GET", headers={"Host": "custom.com"})
        assert request.headers["Host"] == "custom.com"

    def test_request_host_header_case_insensitive_check(self):
        url = URL("http://example.com/path")
        request = Request(url, "GET", headers={"host": "custom.com"})
        assert request.headers.get("Host") == "custom.com"
        host_values = request.headers.getall("Host")
        assert len(host_values) == 1

    def test_request_accept_header_default(self):
        url = URL("http://example.com/path")
        request = Request(url, "GET")
        assert request.headers["Accept"] == "*/*"

    def test_request_accept_header_not_overridden(self):
        url = URL("http://example.com/path")
        request = Request(url, "GET", headers={"Accept": "application/json"})
        assert request.headers["Accept"] == "application/json"

    def test_request_user_agent_header_default(self):
        url = URL("http://example.com/path")
        request = Request(url, "GET")
        assert "User-Agent" in request.headers
        assert "zapros" in request.headers["User-Agent"].lower()

    def test_request_user_agent_header_not_overridden(self):
        url = URL("http://example.com/path")
        request = Request(url, "GET", headers={"User-Agent": "custom-agent"})
        assert request.headers["User-Agent"] == "custom-agent"

    def test_request_accept_encoding_header_default(self):
        url = URL("http://example.com/path")
        request = Request(url, "GET")
        assert "Accept-Encoding" in request.headers

    def test_request_accept_encoding_header_not_overridden(self):
        url = URL("http://example.com/path")
        request = Request(url, "GET", headers={"Accept-Encoding": "gzip"})
        assert request.headers["Accept-Encoding"] == "gzip"

    def test_request_is_replayable_none_body(self):
        url = URL("http://example.com/path")
        request = Request(url, "GET")
        assert request.is_replayable() is True

    def test_request_is_replayable_bytes_body(self):
        url = URL("http://example.com/path")
        request = Request(url, "POST", body=b"data")
        assert request.is_replayable() is True

    def test_request_is_replayable_sync_stream(self):
        url = URL("http://example.com/path")
        request = Request(url, "POST", body=iter([b"chunk"]))
        assert request.is_replayable() is False

    def test_request_is_replayable_async_stream(self):
        url = URL("http://example.com/path")

        async def body_gen():
            yield b"chunk"

        request = Request(url, "POST", body=body_gen())
        assert request.is_replayable() is False

    def test_request_context_default(self):
        url = URL("http://example.com/path")
        request = Request(url, "GET")
        assert request.context == {}

    def test_request_context_provided(self):
        url = URL("http://example.com/path")
        context: RequestContext = {"timeouts": {"connect": 5.0}}
        request = Request(url, "GET", context=context)
        assert request.context == context


class TestResponse:
    def test_response_minimal(self):
        response = Response(200)
        assert response.status == 200
        assert len(response.headers) == 0
        assert response.content is None
        assert response.context == {}

    def test_response_with_headers_dict(self):
        response = Response(200, headers={"Content-Type": "application/json"})
        assert response.headers["Content-Type"] == "application/json"

    def test_response_with_headers_list(self):
        response = Response(200, headers=[("Content-Type", "application/json"), ("Accept", "*/*")])
        assert response.headers["Content-Type"] == "application/json"
        assert response.headers["Accept"] == "*/*"

    def test_response_with_headers_instance(self):
        headers = Headers({"Content-Type": "application/json"})
        response = Response(200, headers=headers)
        assert response.headers["Content-Type"] == "application/json"

    def test_response_content_bytes(self):
        content = b"Hello, World!"
        response = Response(200, content=content)
        assert response.content == content

    def test_response_content_sync_stream(self):
        stream = StreamWrapper(iter([b"chunk1", b"chunk2"]))
        response = Response(200, content=stream)
        assert response.content == stream

    def test_response_content_async_stream(self):
        async def gen():
            yield b"chunk1"

        stream = AsyncStreamWrapper(gen())
        response = Response(200, content=stream)
        assert response.content == stream

    def test_response_text_parameter(self):
        response = Response(200, text="Hello, World!")
        assert response.content == b"Hello, World!"
        assert response.headers["Content-Type"] == "text/plain; charset=utf-8"
        assert "Content-Encoding" not in response.headers

    def test_response_text_strips_content_encoding(self):
        response = Response(200, headers={"Content-Encoding": "gzip"}, text="Hello, World!")
        assert "Content-Encoding" not in response.headers

    def test_response_text_does_not_override_content_type(self):
        response = Response(200, headers={"Content-Type": "custom"}, text="Hello, World!")
        assert response.headers["Content-Type"] == "custom"

    def test_response_json_data_parameter(self):
        data = {"key": "value"}
        response = Response(200, json=data)
        assert isinstance(response.content, bytes)
        assert json_module.loads(response.content) == data
        assert response.headers["Content-Type"] == "application/json; charset=utf-8"
        assert "Content-Encoding" not in response.headers

    def test_response_json_data_strips_content_encoding(self):
        response = Response(200, headers={"Content-Encoding": "gzip"}, json={"key": "value"})
        assert "Content-Encoding" not in response.headers

    def test_response_multiple_body_parameters_raises_valueerror(self):
        with pytest.raises(ValueError) as exc_info:
            Response(200, content=b"data", text="text")  # type: ignore
        assert "Only one of" in str(exc_info.value)

    def test_response_context_default(self):
        response = Response(200)
        assert response.context == {}

    def test_response_context_provided(self):
        context: ResponseContext = {"caching": {"from_cache": True}}
        response = Response(200, context=context)
        assert response.context == context

    def test_response_read_bytes(self):
        response = Response(200, content=b"Hello, World!")
        content = response.read()
        assert content == b"Hello, World!"

    def test_response_read_sync_stream(self):
        stream = StreamWrapper(iter([b"chunk1", b"chunk2"]))
        response = Response(200, content=stream)
        content = response.read()
        assert content == b"chunk1chunk2"

    def test_response_read_none_content(self):
        response = Response(200)
        content = response.read()
        assert content == b""

    def test_response_read_caches_content(self):
        stream = StreamWrapper(iter([b"chunk1", b"chunk2"]))
        response = Response(200, content=stream)
        content1 = response.read()
        content2 = response.read()
        assert content1 == content2
        assert content1 is content2

    @pytest.mark.asyncio
    async def test_response_aread_bytes(self):
        response = Response(200, content=b"Hello, World!")
        content = await response.aread()
        assert content == b"Hello, World!"

    @pytest.mark.asyncio
    async def test_response_aread_async_stream(self):
        async def gen():
            yield b"chunk1"
            yield b"chunk2"

        stream = AsyncStreamWrapper(gen())
        response = Response(200, content=stream)
        content = await response.aread()
        assert content == b"chunk1chunk2"

    @pytest.mark.asyncio
    async def test_response_aread_none_content(self):
        response = Response(200)
        content = await response.aread()
        assert content == b""

    @pytest.mark.asyncio
    async def test_response_aread_caches_content(self):
        async def gen():
            yield b"chunk1"
            yield b"chunk2"

        stream = AsyncStreamWrapper(gen())
        response = Response(200, content=stream)
        content1 = await response.aread()
        content2 = await response.aread()
        assert content1 == content2
        assert content1 is content2

    def test_response_text_method(self):
        response = Response(200, content=b"Hello, World!")
        text = response.text
        assert text == "Hello, World!"

    def test_response_text_with_encoding(self):
        response = Response(
            200,
            headers={"Content-Type": "text/plain; charset=latin-1"},
            content="Hëllo".encode("latin-1"),
        )
        text = response.text
        assert text == "Hëllo"

    def test_response_json_method(self):
        data = {"key": "value"}
        response = Response(200, content=json_module.dumps(data).encode("utf-8"))
        result = response.json
        assert result == data

    def test_response_encoding_from_content_type(self):
        response = Response(200, headers={"Content-Type": "text/html; charset=latin-1"})
        assert response.encoding == "latin-1"

    def test_response_encoding_default_utf8(self):
        response = Response(200)
        assert response.encoding == "utf-8"

    def test_response_encoding_no_content_type(self):
        response = Response(200, headers={})
        assert response.encoding == "utf-8"

    def test_response_iter_bytes_from_bytes(self):
        response = Response(200, content=b"Hello, World!")
        chunks = list(response.iter_bytes())
        result = b"".join(chunks)
        assert result == b"Hello, World!"

    def test_response_iter_bytes_from_stream(self):
        stream = StreamWrapper(iter([b"chunk1", b"chunk2"]))
        response = Response(200, content=stream)
        chunks = list(response.iter_bytes())
        result = b"".join(chunks)
        assert result == b"chunk1chunk2"

    def test_response_iter_bytes_none_content(self):
        response = Response(200)
        chunks = list(response.iter_bytes())
        assert chunks == []

    def test_response_iter_bytes_custom_chunk_size(self):
        response = Response(200, content=b"Hello, World!")
        chunks = list(response.iter_bytes(chunk_size=5))
        assert all(len(chunk) <= 5 for chunk in chunks)

    def test_response_iter_bytes_gzip_decompression(self):
        original = b"Hello, World!"
        compressed = gzip.compress(original)
        stream = StreamWrapper(iter([compressed]))
        response = Response(200, headers={"Content-Encoding": "gzip"}, content=stream)
        chunks = list(response.iter_bytes())
        result = b"".join(chunks)
        assert result == original

    def test_response_iter_bytes_deflate_decompression(self):
        original = b"Hello, World!"
        compressed = zlib.compress(original)
        stream = StreamWrapper(iter([compressed]))
        response = Response(200, headers={"Content-Encoding": "deflate"}, content=stream)
        chunks = list(response.iter_bytes())
        result = b"".join(chunks)
        assert result == original

    def test_response_iter_bytes_multiple_encodings(self):
        original = b"Hello, World!"
        compressed_gzip = gzip.compress(original)
        compressed_deflate = zlib.compress(compressed_gzip)
        stream = StreamWrapper(iter([compressed_deflate]))
        response = Response(200, headers={"Content-Encoding": "gzip, deflate"}, content=stream)
        chunks = list(response.iter_bytes())
        result = b"".join(chunks)
        assert result == original

    def test_response_iter_bytes_case_insensitive_encoding(self):
        original = b"Hello, World!"
        compressed = gzip.compress(original)
        stream = StreamWrapper(iter([compressed]))
        response = Response(200, headers={"Content-Encoding": "GZIP"}, content=stream)
        chunks = list(response.iter_bytes())
        result = b"".join(chunks)
        assert result == original

    def test_response_iter_bytes_unknown_encoding(self):
        original = b"Hello, World!"
        stream = StreamWrapper(iter([original]))
        response = Response(200, headers={"Content-Encoding": "unknown"}, content=stream)
        chunks = list(response.iter_bytes())
        result = b"".join(chunks)
        assert result == original

    def test_response_iter_bytes_identity_encoding(self):
        original = b"Hello, World!"
        stream = StreamWrapper(iter([original]))
        response = Response(200, headers={"Content-Encoding": "identity"}, content=stream)
        chunks = list(response.iter_bytes())
        result = b"".join(chunks)
        assert result == original

    def test_response_iter_bytes_async_stream_raises_typeerror(self):
        async def gen():
            yield b"chunk"

        stream = AsyncStreamWrapper(gen())
        response = Response(200, content=stream)
        with pytest.raises(AsyncSyncMismatchError) as exc_info:
            list(response.iter_bytes())
        assert "using `async_iter_bytes`" in str(exc_info.value)

    @pytest.mark.asyncio
    async def test_response_async_iter_bytes_from_bytes(self):
        response = Response(200, content=b"Hello, World!")
        chunks = [chunk async for chunk in response.async_iter_bytes()]
        result = b"".join(chunks)
        assert result == b"Hello, World!"

    @pytest.mark.asyncio
    async def test_response_async_iter_bytes_from_async_stream(self):
        async def gen():
            yield b"chunk1"
            yield b"chunk2"

        stream = AsyncStreamWrapper(gen())
        response = Response(200, content=stream)
        chunks = [chunk async for chunk in response.async_iter_bytes()]
        result = b"".join(chunks)
        assert result == b"chunk1chunk2"

    @pytest.mark.asyncio
    async def test_response_async_iter_bytes_none_content(self):
        response = Response(200)
        chunks = [chunk async for chunk in response.async_iter_bytes()]
        assert chunks == []

    @pytest.mark.asyncio
    async def test_response_async_iter_bytes_custom_chunk_size(self):
        response = Response(200, content=b"Hello, World!")
        chunks = [chunk async for chunk in response.async_iter_bytes(chunk_size=5)]
        assert all(len(chunk) <= 5 for chunk in chunks)

    @pytest.mark.asyncio
    async def test_response_async_iter_bytes_gzip_decompression(self):
        original = b"Hello, World!"
        compressed = gzip.compress(original)

        async def gen():
            yield compressed

        stream = AsyncStreamWrapper(gen())
        response = Response(200, headers={"Content-Encoding": "gzip"}, content=stream)
        chunks = [chunk async for chunk in response.async_iter_bytes()]
        result = b"".join(chunks)
        assert result == original

    @pytest.mark.asyncio
    async def test_response_async_iter_bytes_sync_stream_raises_typeerror(self):
        stream = StreamWrapper(iter([b"chunk"]))
        response = Response(200, content=stream)
        with pytest.raises(AsyncSyncMismatchError) as exc_info:
            async for _ in response.async_iter_bytes():
                pass
        assert "using `iter_bytes`" in str(exc_info.value)

    def test_response_iter_raw_from_bytes(self):
        response = Response(200, content=b"Hello, World!")
        chunks = list(response.iter_raw())
        result = b"".join(chunks)
        assert result == b"Hello, World!"

    def test_response_iter_raw_from_stream(self):
        original = b"Hello, World!"
        compressed = gzip.compress(original)
        stream = StreamWrapper(iter([compressed]))
        response = Response(200, headers={"Content-Encoding": "gzip"}, content=stream)
        chunks = list(response.iter_raw())
        result = b"".join(chunks)
        assert result == compressed

    def test_response_iter_raw_none_content(self):
        response = Response(200)
        chunks = list(response.iter_raw())
        assert chunks == []

    def test_response_iter_raw_async_stream_raises_typeerror(self):
        async def gen():
            yield b"chunk"

        stream = AsyncStreamWrapper(gen())
        response = Response(200, content=stream)
        with pytest.raises(AsyncSyncMismatchError) as exc_info:
            list(response.iter_raw())
        assert "using `async_iter_raw`" in str(exc_info.value)

    @pytest.mark.asyncio
    async def test_response_async_iter_raw_from_bytes(self):
        response = Response(200, content=b"Hello, World!")
        chunks = [chunk async for chunk in response.async_iter_raw()]
        result = b"".join(chunks)
        assert result == b"Hello, World!"

    @pytest.mark.asyncio
    async def test_response_async_iter_raw_from_async_stream(self):
        original = b"Hello, World!"
        compressed = gzip.compress(original)

        async def gen():
            yield compressed

        stream = AsyncStreamWrapper(gen())
        response = Response(200, headers={"Content-Encoding": "gzip"}, content=stream)
        chunks = [chunk async for chunk in response.async_iter_raw()]
        result = b"".join(chunks)
        assert result == compressed

    @pytest.mark.asyncio
    async def test_response_async_iter_raw_none_content(self):
        response = Response(200)
        chunks = [chunk async for chunk in response.async_iter_raw()]
        assert chunks == []

    @pytest.mark.asyncio
    async def test_response_async_iter_raw_sync_stream_raises_typeerror(self):
        stream = StreamWrapper(iter([b"chunk"]))
        response = Response(200, content=stream)
        with pytest.raises(AsyncSyncMismatchError) as exc_info:
            async for _ in response.async_iter_raw():
                pass
        assert "using `iter_raw`" in str(exc_info.value)

    def test_response_iter_text_from_bytes(self):
        response = Response(200, content=b"Hello, World!")
        chunks = list(response.iter_text())
        result = "".join(chunks)
        assert result == "Hello, World!"

    def test_response_iter_text_with_encoding(self):
        response = Response(
            200,
            headers={"Content-Type": "text/plain; charset=latin-1"},
            content="Hëllo".encode("latin-1"),
        )
        chunks = list(response.iter_text())
        result = "".join(chunks)
        assert result == "Hëllo"

    def test_response_iter_text_custom_chunk_size(self):
        response = Response(200, content=b"Hello, World!")
        chunks = list(response.iter_text(chunk_size=5))
        assert len(chunks) > 0

    @pytest.mark.asyncio
    async def test_response_async_iter_text_from_bytes(self):
        response = Response(200, content=b"Hello, World!")
        chunks = [chunk async for chunk in response.async_iter_text()]
        result = "".join(chunks)
        assert result == "Hello, World!"

    def test_response_read_then_iterate(self):
        stream = StreamWrapper(iter([b"chunk1", b"chunk2"]))
        response = Response(200, content=stream)
        content = response.read()
        assert content == b"chunk1chunk2"
        chunks = list(response.iter_bytes())
        result = b"".join(chunks)
        assert result == b"chunk1chunk2"

    def test_response_iterate_then_read(self):
        stream = StreamWrapper(iter([b"chunk1", b"chunk2"]))
        response = Response(200, content=stream)
        chunks = list(response.iter_bytes())
        result = b"".join(chunks)
        assert result == b"chunk1chunk2"
        content = response.read()
        assert content == b"chunk1chunk2"

    def test_response_close_bytes_content(self):
        response = Response(200, content=b"Hello, World!")
        response.close()

    def test_response_close_none_content(self):
        response = Response(200)
        response.close()

    def test_response_close_sync_stream(self):
        stream = StreamWrapper(iter([b"chunk"]))
        response = Response(200, content=stream)
        response.close()
        assert stream._closed is True

    def test_response_close_async_stream_raises_typeerror(self):
        async def gen():
            yield b"chunk"

        stream = AsyncStreamWrapper(gen())
        response = Response(200, content=stream)
        with pytest.raises(AsyncSyncMismatchError) as exc_info:
            response.close()
        assert "use `aclose()`" in str(exc_info.value)

    def test_response_close_after_stream_consumed(self):
        stream = StreamWrapper(iter([b"chunk"]))
        response = Response(200, content=stream)
        response.read()
        response.close()

    def test_response_close_idempotent(self):
        stream = StreamWrapper(iter([b"chunk"]))
        response = Response(200, content=stream)
        response.close()
        response.close()

    @pytest.mark.asyncio
    async def test_response_aclose_bytes_content(self):
        response = Response(200, content=b"Hello, World!")
        await response.aclose()

    @pytest.mark.asyncio
    async def test_response_aclose_none_content(self):
        response = Response(200)
        await response.aclose()

    @pytest.mark.asyncio
    async def test_response_aclose_async_stream(self):
        async def gen():
            yield b"chunk"

        stream = AsyncStreamWrapper(gen())
        response = Response(200, content=stream)
        await response.aclose()
        assert stream._closed is True

    @pytest.mark.asyncio
    async def test_response_aclose_sync_stream_raises_typeerror(self):
        stream = StreamWrapper(iter([b"chunk"]))
        response = Response(200, content=stream)
        with pytest.raises(AsyncSyncMismatchError) as exc_info:
            await response.aclose()
        assert "use `close()`" in str(exc_info.value)

    @pytest.mark.asyncio
    async def test_response_aclose_after_stream_consumed(self):
        async def gen():
            yield b"chunk"

        stream = AsyncStreamWrapper(gen())
        response = Response(200, content=stream)
        await response.aread()
        await response.aclose()

    @pytest.mark.asyncio
    async def test_response_aclose_idempotent(self):
        async def gen():
            yield b"chunk"

        stream = AsyncStreamWrapper(gen())
        response = Response(200, content=stream)
        await response.aclose()
        await response.aclose()

    def test_response_context_manager(self):
        stream = StreamWrapper(iter([b"chunk"]))
        with Response(200, content=stream):
            pass
        assert stream._closed is True

    @pytest.mark.asyncio
    async def test_response_async_context_manager(self):
        async def gen():
            yield b"chunk"

        stream = AsyncStreamWrapper(gen())
        async with Response(200, content=stream):
            pass
        assert stream._closed is True

    def test_response_stream_consumption_caching(self):
        stream = StreamWrapper(iter([b"chunk1", b"chunk2"]))
        response = Response(200, content=stream)
        list(response.iter_bytes())
        assert isinstance(response.content, bytes)
